#include <iostream>
#include <map>

using namespace std;

#include "Clusters.h"
#include "../PPDTable/PPDSampleType.h"

unsigned CLUSTER_NO = 1000;

Clusters::Clusters(VectString *d): data(d) 
{
  clusters=new VectCluster;
  ppdtable=new PPDTable();
  query_pivot = new ContQueryPivot();
}

Clusters::Clusters(const Clusters& cs): data(cs.data)
{
  clusters=new VectCluster;
  *clusters=*cs.clusters;
  ppdtable=new PPDTable();
  *ppdtable=*cs.ppdtable;
  query_pivot = new ContQueryPivot();
  *query_pivot = *cs.query_pivot;
}

Clusters::~Clusters() 
{
  delete clusters;
  delete ppdtable;
  delete query_pivot;
}

Clusters& Clusters::operator=(const Clusters& cs) 
{
  if (this==&cs)
    return *this;
  *clusters=*cs.clusters;
  *ppdtable=*cs.ppdtable;
  *query_pivot = *cs.query_pivot;
  data=cs.data;
  return *this;
}

void Clusters::push_back_cluster(const Cluster& c) 
{
  this->clusters->push_back(c);
}

const VectClusterIt Clusters::begin_cluster() const 
{
  return clusters->begin();
}

const VectClusterIt Clusters::end_cluster() const 
{
  return clusters->end();
}

const size_t Clusters::size_cluster() const 
{
  return clusters->size();
}

const Cluster& Clusters::get_cluster(unsigned i) const 
{
  return (*clusters)[i];
}

PPDTable& Clusters::get_ppdtable() const 
{
  return *ppdtable;
}

ContQueryPivot& Clusters::get_query_pivot() const 
{
  return *query_pivot;
}

const ContPPDEntryIt Clusters::begin_ppdtable() const 
{
  return ppdtable->begin_table();
}

const ContPPDEntryIt Clusters::end_ppdtable() const 
{
  return ppdtable->end_table();
}

const ContPPDEntryIt Clusters::find_ppdtable(const PPDEntry& e) const 
{
  return ppdtable->find_table(e);
}

const ContPPDEntryIt Clusters::upper_bound_ppdtable(const PPDEntry& e) const 
{
  return ppdtable->upper_bound_table(e);
}

const ContPPDEntryIt Clusters::lower_bound_ppdtable(const PPDEntry& e) const 
{
  return ppdtable->lower_bound_table(e);
}

void Clusters::build_ftables() 
{
  for (VectClusterIt i=clusters->begin(); i!=clusters->end(); i++)
    i->build_ftable(data);
}

void Clusters::build_ppdtable() 
{
  PPDSampleType sample_engine=PPDSampleType(data, this, query_pivot);

  while (sample_engine.has_next()) {
    PPDTriple triple=sample_engine.next();
    PPDEntry entry=PPDEntry(triple.vect1, triple.vect2, UpperOrEqual(triple.dist));
    ppdtable->insert(entry);   
  }
}

ostream& operator<<(ostream& out, const Clusters& cs) 
{
  out << static_cast<unsigned>(cs.clusters->size()) << endl << endl;
  copy(cs.clusters->begin(), cs.clusters->end(), ostream_iterator<Cluster>(out, "\n"));
  return out;
}

istream& operator>>(istream& in, Clusters& cs) 
{
  unsigned n;
  in>>n;
  for (unsigned i=0; i<n; i++) {
    Cluster cluster;			
    in>>cluster;
    cs.clusters->push_back(cluster);
  }
  return in;
}

typedef map<SimType, unsigned> MapSimDist;

ostream& operator<<(ostream& out, const MapSimDist& map) 
{
  for (MapSimDist::const_iterator it = map.begin(); it != map.end(); ++it) 
    {
      out << it->first << " " << it->second << endl;
    }
  return out;
}

void Clusters::hist(ostream &out, const VectString& d) const 
{
  MapSimDist global_hist, local_hist;
  for (VectClusterIt cluster = clusters->begin(); cluster!= clusters->end(); ++cluster) 
    {
      local_hist.clear();
      unsigned pivot = cluster->get_pivot();
      for (ContCluster::const_iterator index = cluster->begin(); index != cluster->end(); ++index)
        {
          SimType dist = SimDist(d[pivot], d[*index]);	  
          ++local_hist[UpperOrEqual(dist)];
        }

      for (MapSimDist::const_iterator it = local_hist.begin(); it != local_hist.end(); ++it)
        {
          global_hist[it->first] += it->second; // / cluster->size();
        }      
    }

  out << global_hist;
}

void Clusters::stat(ostream &out, const VectString& d) const 
{
  float 
    sum_avg = 0, 
    sum_rad = 0;
  for (VectClusterIt cluster = clusters->begin(); cluster!= clusters->end(); ++cluster) 
    {
      unsigned pivot = cluster->get_pivot();
      SimType
        sum_dist = 0;
      for (ContCluster::const_iterator index = cluster->begin(); index != cluster->end(); ++index)
        {
          SimType dist = SimDist(d[pivot], d[*index]);	  
          sum_dist += dist;
        }
      SimType
        avg_dist = sum_dist / cluster->size();
      sum_avg += avg_dist;
      sum_rad += cluster->get_radius();
    }

  out << sum_avg / size_cluster() << "\t" << sum_rad / size_cluster() << "\t";
}

ostream& Clusters::info(ostream& out) 
{
  out << "Clusters" << endl << "---" << endl;
  out << "Clusters no\t" << CLUSTER_NO << endl;
  return out << endl;
}

bool Clusters::is_pivot(unsigned idx) const
{
  for (vector<Cluster>::const_iterator it = clusters->begin(); it != clusters->end(); ++it) 
    {
      if (it->get_pivot() == idx) 
        {
          return true;
        }
    }
  return false;
}

void Clusters::insert(unsigned idx)
{
  SimType min = SimDist((*data)[clusters->begin()->get_pivot()], (*data)[idx]);
  vector<Cluster>::iterator closest = clusters->begin();
  for (vector<Cluster>::iterator it = clusters->begin() + 1; it != clusters->end(); ++it) 
    {
      SimType crt = SimDist((*data)[it->get_pivot()], (*data)[idx]);
      if (crt < min) 
        {
          min = crt;
          closest = it;
        }
    }
  //   cerr << "...closest\t" << closest->get_pivot() << "\t" << min << endl;

  // "closest" is the closest cluster
  unsigned pivot = closest->get_pivot();
  SimVect vect2 = SimVect( (*data)[pivot], (*data)[idx]);
  //   cerr << "...insert ftable\t" << pivot << "\t" << vect2 << endl;
  closest->insert(idx, vect2);

  SimType dist = vect2.get_dist();
  if (dist > closest->get_radius()) 
    {
      closest->set_radius(dist);
    }

  for (ContQueryPivot::const_iterator it = query_pivot->find(pivot);
       it != query_pivot->end() && it->first == pivot;
       ++it) 
    {
      SimVect vect1 = it->second.vect;
      SimType dist = SimDist((*data)[it->second.query], (*data)[idx]);
      PPDEntry entry = PPDEntry(vect1, vect2, UpperOrEqual(dist));
      //       cerr << "...insert ppdtable\t" << entry << endl;
      ppdtable->insert(entry);
    }
}

class CmpDist
{
private:
  const vector<icstring> *data;
  icstring pivot_str;
public:
  CmpDist(const vector<icstring> *data, unsigned pivot): data(data), pivot_str((*data)[pivot])
  {
  }
  bool operator()(unsigned i, unsigned j)
  {
    return SimDist((*data)[i], pivot_str) < SimDist((*data)[j], pivot_str);
  }
};

void Clusters::erase(unsigned idx) 
{
  vector<Cluster>::iterator contain;
  for (contain = clusters->begin(); contain != clusters->end(); ++contain) 
    {
      if (contain->find(idx) != contain->end()) 
        {
          break;
        }
    }
  if (contain == clusters->end())
    {
      cerr << "Clusters::erase index not found\t" << idx << endl;
      return;
    }  
  unsigned pivot = contain->get_pivot();
  SimVect vect2 = SimVect((*data)[pivot], (*data)[idx]);
  //   cerr << "...erase ftable\t" << pivot << "\t" << vect2 << endl;
  contain->erase(idx, vect2);

  if (contain->get_radius() == vect2.get_dist())
    {
      // update radius
      contain->set_radius(SimDist((*data)[pivot], 
                                  (*data)[*max_element(contain->begin(), 
                                                       contain->end(),
                                                       CmpDist(data, pivot))]));
    }

  for (ContQueryPivot::const_iterator it = query_pivot->find(pivot);
       it != query_pivot->end() && it->first == pivot;
       ++it) 
    {
      SimVect vect1 = it->second.vect;
      SimType dist = SimDist((*data)[it->second.query], (*data)[idx]);
      PPDEntry entry = PPDEntry(vect1, vect2, UpperOrEqual(dist));
      //       cerr << "...erase ppdtable\t" << entry << endl;
      //       cerr << "query_pivot\t" << it->first << "\t" << it->second << endl;
      ppdtable->erase(entry);
    }
}

ostream& operator<<(ostream& out, const ContQueryPivot& cont) 
{
  out << cont.size() << endl << endl;
  for(ContQueryPivot::const_iterator it = cont.begin(); it != cont.end(); ++it) 
    {
      out << it->first << "\t" << it->second << endl;
    }
  return out;
}

istream& operator>>(istream& in, ContQueryPivot& cont) 
{
  unsigned sz;
  in >> sz;
  for (unsigned i = 0; i < sz; i++)
    {
      unsigned pivot;
      QueryVect qv;
      in >> pivot >> qv;
      cont.insert(std::make_pair(pivot, qv));
    }
  return in;
}
